import sys

import vowpalwabbit

# Known limitations:
# - Stdin cannot be used
# - vw.exe specifc args cannot be used: --args, --onethread


def main():
    opts = sys.argv[1:]
    vowpalwabbit.Workspace(" ".join(opts))


if __name__ == "__main__":
    main()



import pandas as pd
from os import path
import vowpalwabbit

import unittest
import platform


def helper_get_test_dir():
    curr_path = path.dirname(path.realpath(__file__))
    return path.join(path.dirname(path.dirname(curr_path)), "test")


def helper_get_data():
    train_data = [
        {
            "action": 1,
            "cost": 2,
            "probability": 0.4,
            "feature1": "a",
            "feature2": "c",
            "feature3": "",
        },
        {
            "action": 3,
            "cost": 0,
            "probability": 0.2,
            "feature1": "b",
            "feature2": "d",
            "feature3": "",
        },
        {
            "action": 4,
            "cost": 1,
            "probability": 0.5,
            "feature1": "a",
            "feature2": "b",
            "feature3": "",
        },
        {
            "action": 2,
            "cost": 1,
            "probability": 0.3,
            "feature1": "a",
            "feature2": "b",
            "feature3": "c",
        },
        {
            "action": 3,
            "cost": 1,
            "probability": 0.7,
            "feature1": "a",
            "feature2": "d",
            "feature3": "",
        },
    ]

    train_df = pd.DataFrame(train_data)

    train_df["index"] = range(1, len(train_df) + 1)
    train_df = train_df.set_index("index")

    test_data = [
        {"feature1": "b", "feature2": "c", "feature3": ""},
        {"feature1": "a", "feature2": "", "feature3": "b"},
        {"feature1": "b", "feature2": "b", "feature3": ""},
        {"feature1": "a", "feature2": "", "feature3": "b"},
    ]

    test_df = pd.DataFrame(test_data)

    # Add index to data frame
    test_df["index"] = range(1, len(test_df) + 1)
    test_df = test_df.set_index("index")

    return train_df, test_df


def test_getting_started_example_cb():
    return helper_getting_started_example("--cb")


def test_getting_started_example_legacy_cb():
    return helper_getting_started_example("--cb_force_legacy --cb")


@unittest.skipIf(
    platform.machine() == "aarch64", "skipping due to floating-point error on aarch64"
)
def helper_getting_started_example(which_cb):
    train_df, test_df = helper_get_data()

    vw = vowpalwabbit.Workspace(which_cb + " 4 --log_level off", enable_logging=True)

    for i in train_df.index:
        action = train_df.loc[i, "action"]
        cost = train_df.loc[i, "cost"]
        probability = train_df.loc[i, "probability"]
        feature1 = train_df.loc[i, "feature1"]
        feature2 = train_df.loc[i, "feature2"]
        feature3 = train_df.loc[i, "feature3"]

        learn_example = (
            str(action)
            + ":"
            + str(cost)
            + ":"
            + str(probability)
            + " | "
            + str(feature1)
            + " "
            + str(feature2)
            + " "
            + str(feature3)
        )
        vw.learn(learn_example)

    assert (
        vw.get_prediction_type() == vw.pMULTICLASS
    ), "prediction_type should be multiclass"

    for j in test_df.index:
        feature1 = test_df.loc[j, "feature1"]
        feature2 = test_df.loc[j, "feature2"]
        feature3 = test_df.loc[j, "feature3"]
        choice = vw.predict(
            "| " + str(feature1) + " " + str(feature2) + " " + str(feature3)
        )
        assert isinstance(choice, int), "choice should be int"
        assert choice == 3, "predicted action should be 3 instead of " + str(choice)

    # test that metrics is empty since "--extra_metrics filename" was not supplied
    assert len(vw.get_learner_metrics()) == 0

    vw.finish()

    output = vw.get_log()

    if which_cb.find("legacy") != -1:
        test_file = "test-sets/ref/python_test_cb_legacy.stderr"
    else:
        test_file = "test-sets/ref/python_test_cb.stderr"

    with open(path.join(helper_get_test_dir(), test_file), "r") as file:
        actual = file.readlines()
        for j, i in zip(actual, output):
            assert i == j, "line mismatch should be: " + j + " output: " + i


def test_getting_started_example_with():
    train_df, test_df = helper_get_data()

    # with syntax calls into vw.finish() automatically.
    # you actually want to use 'with vowpalwabbit.Workspace("--cb 4") as vw:'
    # but we need to assert on vw.finished for test purposes
    vw = vowpalwabbit.Workspace("--cb 4")
    with vw as vw:
        for i in train_df.index:
            action = train_df.loc[i, "action"]
            cost = train_df.loc[i, "cost"]
            probability = train_df.loc[i, "probability"]
            feature1 = train_df.loc[i, "feature1"]
            feature2 = train_df.loc[i, "feature2"]
            feature3 = train_df.loc[i, "feature3"]

            learn_example = (
                str(action)
                + ":"
                + str(cost)
                + ":"
                + str(probability)
                + " | "
                + str(feature1)
                + " "
                + str(feature2)
                + " "
                + str(feature3)
            )
            vw.learn(learn_example)

        assert (
            vw.get_prediction_type() == vw.pMULTICLASS
        ), "prediction_type should be multiclass"

        for j in test_df.index:
            feature1 = test_df.loc[j, "feature1"]
            feature2 = test_df.loc[j, "feature2"]
            feature3 = test_df.loc[j, "feature3"]
            choice = vw.predict(
                "| " + str(feature1) + " " + str(feature2) + " " + str(feature3)
            )
            assert isinstance(choice, int), "choice should be int"
            assert choice == 3, "predicted action should be 3"

    assert vw.finished == True, "with syntax should finish() vw instance"



import vowpalwabbit

# Named specifically as the delimiter used is specific for the number of actions
# used in this test case.
def count_weights_from_readable_model_file_for_equiv_test(file_name):
    with open(file_name) as file:
        model_file_contents = file.read()
        DELIM = "action_sum 5\n:0\n"
        weights_idx = model_file_contents.find(DELIM) + len(DELIM)
        return model_file_contents[weights_idx:].count("\n")


def test_ccb_single_slot_and_cb_equivalence_no_slot_features():
    # --- CCB
    ccb_model_file_name = "model_file_ccb_equiv.txt"
    ccb_workspace = vowpalwabbit.Workspace(
        quiet=True,
        predict_only_model=True,
        ccb_explore_adf=True,
        readable_model=ccb_model_file_name,
    )

    ccb_ex = """
    ccb shared |User b
    ccb action |Action d
    ccb action |Action e
    ccb action |Action f
    ccb action |Action ff
    ccb action |Action fff
    ccb slot 4:1:0.2 |
    """
    ccb_workspace.learn(ccb_ex)
    ccb_workspace.finish()

    ccb_num_weights = count_weights_from_readable_model_file_for_equiv_test(
        ccb_model_file_name
    )

    # --- CB
    cb_model_file_name = "model_file_cb_equiv.txt"
    cb_workspace = vowpalwabbit.Workspace(
        quiet=True,
        predict_only_model=True,
        cb_explore_adf=True,
        readable_model=cb_model_file_name,
    )

    cb_ex = """
    shared |User b
    |Action d
    |Action e
    |Action f
    |Action ff
    4:1:0.2 |Action fff
    """

    cb_workspace.learn(cb_ex)
    cb_workspace.finish()
    cb_num_weights = count_weights_from_readable_model_file_for_equiv_test(
        cb_model_file_name
    )

    assert ccb_num_weights == cb_num_weights


def test_ccb_single_slot_and_cb_non_equivalence_with_slot_features():
    # --- CCB
    ccb_model_file_name = "model_file_ccb_no_equiv.txt"
    ccb_workspace = vowpalwabbit.Workspace(
        quiet=True, ccb_explore_adf=True, readable_model=ccb_model_file_name
    )

    ccb_ex = """
    ccb shared |User b
    ccb action |Action d
    ccb action |Action e
    ccb action |Action f
    ccb action |Action ff
    ccb action |Action fff
    ccb slot 4:1:0.2 | slot_feature_1
    """
    ccb_workspace.learn(ccb_ex)
    ccb_workspace.finish()

    ccb_num_weights = count_weights_from_readable_model_file_for_equiv_test(
        ccb_model_file_name
    )

    # --- CB
    cb_model_file_name = "model_file_cb_no_equiv.txt"
    cb_workspace = vowpalwabbit.Workspace(
        quiet=True, cb_explore_adf=True, readable_model=cb_model_file_name
    )

    cb_ex = """
    shared |User b
    |Action d
    |Action e
    |Action f
    |Action ff
    4:1:0.2 |Action fff
    """

    cb_workspace.learn(cb_ex)
    cb_workspace.finish()
    cb_num_weights = count_weights_from_readable_model_file_for_equiv_test(
        cb_model_file_name
    )

    # Since there was at least one slot feature supplied, the equivalent mode
    # does not apply and so we expect there to be more weights in the CCB model.
    assert ccb_num_weights > cb_num_weights


def test_ccb_non_slot_none_outcome():
    model = vowpalwabbit.Workspace(quiet=True, ccb_explore_adf=True)
    example = vowpalwabbit.Example(
        vw=model, labelType=vowpalwabbit.LabelType.CONDITIONAL_CONTEXTUAL_BANDIT
    )
    label = example.get_label(vowpalwabbit.CCBLabel)
    # CCB label is set to UNSET by default.
    assert label.type == vowpalwabbit.CCBLabelType.UNSET
    assert label.outcome is None



class OnlineDRO:
    class OnlineCressieReadLB:
        from math import inf

        @staticmethod
        def intervalimpl(
            n,
            sumw,
            sumwsq,
            sumwr,
            sumwsqr,
            sumwsqrsq,
            wmin,
            wmax,
            alpha=0.05,
            rmin=0,
            rmax=1,
            raiseonerr=False,
        ):
            from math import inf, isclose, sqrt
            from scipy.stats import chi2

            assert wmin < 1
            assert wmax > 1
            assert rmin <= rmax

            uncwfake = wmax if sumw < n else wmin
            if uncwfake == inf:
                uncgstar = 1 + 1 / n
            else:
                unca = (uncwfake + sumw) / (1 + n)
                uncb = (uncwfake ** 2 + sumwsq) / (1 + n)
                uncgstar = (n + 1) * (unca - 1) ** 2 / (uncb - unca * unca)
            Delta = chi2.isf(q=alpha, df=1)
            phi = (-uncgstar - Delta) / (2 * (n + 1))

            bounds = []
            for r, sign in ((rmin, 1),):
                candidates = []
                for wfake in (wmin, wmax):
                    if wfake == inf:
                        x = sign * (r + (sumwr - sumw * r) / n)
                        y = (r * sumw - sumwr) ** 2 / (n * (1 + n)) - (
                            r ** 2 * sumwsq - 2 * r * sumwsqr + sumwsqrsq
                        ) / (1 + n)
                        z = phi + 1 / (2 * n)
                        if isclose(y * z, 0, abs_tol=1e-9):
                            y = 0

                        if z <= 0 and y * z >= 0:
                            kappa = sqrt(y / (2 * z))
                            if isclose(kappa, 0):
                                candidates.append((sign * r, None))
                            else:
                                gstar = x - sqrt(2 * y * z)
                                gamma = (
                                    -kappa * (1 + n) / n + sign * (r * sumw - sumwr) / n
                                )
                                beta = -sign * r
                                candidates.append(
                                    (
                                        gstar,
                                        {
                                            "kappastar": kappa,
                                            "betastar": beta,
                                            "gammastar": gamma,
                                            "wfake": wfake,
                                            # Q_{w,r} &= -\frac{\gamma + \beta w + w r}{(N+1) \kappa} \\
                                            "qfunc": lambda c, w, r, k=kappa, g=gamma, b=beta, s=sign, num=n: -c
                                            * (g + (b + s * r) * w)
                                            / ((num + 1) * k),
                                        },
                                    )
                                )
                    else:
                        barw = (wfake + sumw) / (1 + n)
                        barwsq = (wfake * wfake + sumwsq) / (1 + n)
                        barwr = sign * (wfake * r + sumwr) / (1 + n)
                        barwsqr = sign * (wfake * wfake * r + sumwsqr) / (1 + n)
                        barwsqrsq = (wfake * wfake * r * r + sumwsqrsq) / (1 + n)

                        if barwsq > barw ** 2:
                            x = barwr + (
                                (1 - barw)
                                * (barwsqr - barw * barwr)
                                / (barwsq - barw ** 2)
                            )
                            y = (barwsqr - barw * barwr) ** 2 / (barwsq - barw ** 2) - (
                                barwsqrsq - barwr ** 2
                            )
                            z = phi + (1 / 2) * (1 - barw) ** 2 / (barwsq - barw ** 2)

                            if isclose(y * z, 0, abs_tol=1e-9):
                                y = 0

                            if z <= 0 and y * z >= 0:
                                kappa = sqrt(y / (2 * z)) if y * z > 0 else 0
                                if isclose(kappa, 0):
                                    candidates.append((sign * r, None))
                                else:
                                    gstar = x - sqrt(2 * y * z)
                                    beta = (
                                        -kappa * (1 - barw) - (barwsqr - barw * barwr)
                                    ) / (barwsq - barw * barw)
                                    gamma = -kappa - beta * barw - barwr

                                    candidates.append(
                                        (
                                            gstar,
                                            {
                                                "kappastar": kappa,
                                                "betastar": beta,
                                                "gammastar": gamma,
                                                "wfake": wfake,
                                                # Q_{w,r} &= -\frac{\gamma + \beta w + w r}{(N+1) \kappa} \\
                                                "qfunc": lambda c, w, r, k=kappa, g=gamma, b=beta, s=sign, num=n: -c
                                                * (g + (b + s * r) * w)
                                                / ((num + 1) * k),
                                            },
                                        )
                                    )

                best = min(candidates, key=lambda x: x[0])
                vbound = min(rmax, max(rmin, sign * best[0]))
                bounds.append((vbound, best[1]))

            return (bounds[0][0],), (bounds[0][1],)

        def __init__(self, alpha, tau=1, wmin=0, wmax=inf):
            import numpy as np

            self.alpha = alpha
            self.tau = tau
            self.n = 0
            self.sumw = 0
            self.sumwsq = 0
            self.sumwr = 0
            self.sumwsqr = 0
            self.sumwsqrsq = 0
            self.wmin = wmin
            self.wmax = wmax

            self.duals = None
            self.mleduals = None

        def update(self, c, w, r):
            if c > 0:
                assert (
                    w + 1e-6 >= self.wmin and w <= self.wmax + 1e-6
                ), "w = {} < {} < {}".format(self.wmin, w, self.wmax)
                assert r >= 0 and r <= 1, "r = {}".format(r)

                decay = self.tau ** c
                self.n = decay * self.n + c
                self.sumw = decay * self.sumw + c * w
                self.sumwsq = decay * self.sumwsq + c * w ** 2
                self.sumwr = decay * self.sumwr + c * w * r
                self.sumwsqr = decay * self.sumwsqr + c * (w ** 2) * r
                self.sumwsqrsq = decay * self.sumwsqrsq + c * (w ** 2) * (r ** 2)

                self.duals = None
                self.mleduals = None

            return self

        def recomputeduals(self):
            from crminustwo import CrMinusTwo as CrMinusTwo

            self.duals = self.intervalimpl(
                self.n,
                self.sumw,
                self.sumwsq,
                self.sumwr,
                self.sumwsqr,
                self.sumwsqrsq,
                self.wmin,
                self.wmax,
                self.alpha,
                raiseonerr=True,
            )

        def qlb(self, w, r):
            if self.duals is None:
                self.recomputeduals()

                assert self.duals is not None

            return (
                self.duals[1][0]["qfunc"](1, w, r)
                if self.duals[1][0] is not None
                else 1
            )

    def recompute_duals_test():
        from math import exp, pi
        from pprint import pformat
        import numpy

        ocrl = OnlineDRO.OnlineCressieReadLB(alpha=0.05, tau=0.999)

        ws = numpy.random.RandomState(seed=42).exponential(size=10)
        rs = numpy.random.RandomState(seed=2112).random_sample(size=10)
        duals = []
        bounds = []

        print(list(zip(ws, rs)))

        for (w, r) in zip(ws, rs):
            ocrl.update(1, w, r)
            ocrl.recomputeduals()
            if ocrl.duals[1][0] is None:
                duals.append((True, 0, 0, 0, 0))
            else:
                bounds.append(ocrl.duals[0][0])
                duals.append(
                    (
                        False,
                        ocrl.duals[1][0]["kappastar"],
                        ocrl.duals[1][0]["gammastar"],
                        ocrl.duals[1][0]["betastar"],
                        ocrl.n,
                    )
                )

        print(pformat(bounds))
        print(pformat(duals))

    def qlb_test():
        from math import exp, pi
        from pprint import pformat
        import numpy

        ocrl = OnlineDRO.OnlineCressieReadLB(alpha=0.05, tau=0.999)

        ws = numpy.random.RandomState(seed=42).exponential(size=10)
        rs = numpy.random.RandomState(seed=2112).random_sample(size=10)
        qlbs = []

        print(list(zip(ws, rs)))

        for (w, r) in zip(ws, rs):
            ocrl.update(1, w, r)
            qlbs.append(ocrl.qlb(w, r))

        print(pformat(qlbs))


def main():
    print("generating unit test data. see distributionally_robust_test.cc")
    OnlineDRO.recompute_duals_test()
    OnlineDRO.qlb_test()


if __name__ == "__main__":
    main()



class CrMinusTwo:
    @staticmethod
    def estimate(datagen, wmin, wmax, rmin=0, rmax=1, raiseonerr=False, censored=False):
        from math import inf

        n, sumw, sumwsq, sumwr, sumwsqr, sumwany, sumwsqany = 0, 0, 0, 0, 0, 0, 0
        for c, w, r in datagen():
            n += c
            sumw += c * w
            sumwsq += c * w * w
            if r is not None:
                sumwany += c * w
                sumwsqany += c * w * w
                sumwr += c * w * r
                sumwsqr += c * w * w * r

        assert n > 0

        return CrMinusTwo.estimateimpl(
            n,
            sumw,
            sumwsq,
            sumwr,
            sumwsqr,
            sumwany,
            sumwsqany,
            wmin,
            wmax,
            rmin,
            rmax,
            raiseonerr,
            censored,
        )

    def estimateimpl(
        n,
        sumw,
        sumwsq,
        sumwr,
        sumwsqr,
        sumwany,
        sumwsqany,
        wmin,
        wmax,
        rmin=0,
        rmax=1,
        raiseonerr=False,
        censored=False,
    ):
        from math import inf

        assert wmin >= 0
        assert wmin < 1
        assert wmax > 1
        assert rmax >= rmin

        wfake = wmax if sumw < n else wmin

        if wfake == inf:
            gammastar = -(1 + n) / n
            betastar = 0
            gstar = 1 + 1 / n
        else:
            a = (wfake + sumw) / (1 + n)
            b = (wfake ** 2 + sumwsq) / (1 + n)
            assert a * a < b
            gammastar = (b - a) / (a * a - b)
            betastar = (1 - a) / (a * a - b)
            gstar = (n + 1) * (a - 1) ** 2 / (b - a * a)

        vhat = (-gammastar * sumwr - betastar * sumwsqr) / (1 + n)
        missing = max(0, 1 - (-gammastar * sumw - betastar * sumwsq) / (1 + n))

        if censored:
            # vhat = E[w r 1_{r is not None}] / E[w 1_{r is not None}]

            vnumhat = vhat
            vdenomhat = (-gammastar * sumwany - betastar * sumwsqany) / (1 + n)

            # Minimize[{ (x + a r) / (y + a),  y >= 0, 0 <= a <= m }, a]
            #
            # extrema is always at endpoints of interval

            vmincandidates = []
            vmaxcandidates = []
            if vdenomhat > 0:
                vmincandidates.append(vnumhat / vdenomhat)
                vmaxcandidates.append(vnumhat / vdenomhat)

            if vdenomhat + missing > 0:
                vmincandidates.append(
                    (vnumhat + missing * rmin) / (vdenomhat + missing)
                )
                vmaxcandidates.append(
                    (vnumhat + missing * rmax) / (vdenomhat + missing)
                )

            vmin = min(vmincandidates, default=None)
            vmax = max(vmaxcandidates, default=None)
            vhat = None if vmin is None or vmax is None else (vmin + vmax) / 2
        else:
            vmin = vhat + missing * rmin
            vmax = vhat + missing * rmax
            vhat += missing * (rmin + rmax) / 2

        vmin, vmax, vhat = (
            None if x is None else min(rmax, max(rmin, x)) for x in (vmin, vmax, vhat)
        )

        return vhat, {
            "primal": gstar,
            "gammastar": gammastar,
            "betastar": betastar,
            "vmin": vmin,
            "vmax": vmax,
            "num": n,
            "qfunc": lambda c, w, r: (c / (1 + n)) * (-gammastar - betastar * w),
        }

    @staticmethod
    def estimatediff(
        datagen,
        umin,
        umax,
        wmin,
        wmax,
        rmin=0,
        rmax=1,
        raiseonerr=False,
        censored=False,
    ):
        import numpy as np

        assert umin >= 0
        assert umin < 1
        assert umax > 1
        assert wmin >= 0
        assert wmin < 1
        assert wmax > 1
        assert rmax >= rmin
        assert not censored

        n, sumu, sumw, sumuw, sumusq, sumwsq = 0, 0, 0, 0, 0, 0
        sumuMwr, sumuuMwr, sumwuMwr = 0, 0, 0
        for c, u, w, r in datagen():
            n += c
            sumu += c * u
            sumw += c * w
            sumuw += c * u * w
            sumusq += c * u ** 2
            sumwsq += c * w ** 2
            sumuMwr += c * (u - w) * r
            sumuuMwr += c * u * (u - w) * r
            sumwuMwr += c * w * (u - w) * r

        assert n > 0

        ufake = umax if sumu < n else umin
        wfake = wmax if sumw < n else wmin

        ubar = (sumu + ufake) / (n + 1)
        usqbar = (sumusq + ufake ** 2) / (n + 1)
        uwbar = (sumuw + ufake * wfake) / (n + 1)
        wbar = (sumw + wfake) / (n + 1)
        wsqbar = (sumwsq + wfake ** 2) / (n + 1)

        A = np.array(
            [[-1, -ubar, -wbar], [-ubar, -usqbar, -uwbar], [-wbar, -uwbar, -wsqbar]],
            dtype="float64",
        )
        b = np.ones(3, dtype="float64")

        xstar = np.linalg.lstsq(A, b, rcond=-1)[0]
        beta, gamma, tau = xstar

        deltavhat = (-beta * sumuMwr - gamma * sumuuMwr - tau * sumwuMwr) / n
        missing = (
            -beta * (ufake - wfake)
            - gamma * (ufake ** 2 - ufake * wfake)
            - tau * (ufake * wfake - wfake ** 2)
        ) / (n + 1)

        deltavmin = deltavhat + min(rmin * missing, rmax * missing)
        deltavmax = deltavhat + max(rmin * missing, rmax * missing)
        deltavhat = (deltavmin + deltavmax) / 2

        deltavmin, deltavmax, deltavhat = (
            min(rmax - rmin, max(rmin - rmax, x))
            for x in (deltavmin, deltavmax, deltavhat)
        )

        qfunc = lambda c, u, w, r, n=n, b=beta, g=gamma, t=tau: (c / (n + 1)) * (
            -b - g * u - t * w
        )

        return deltavhat, {
            "deltavmin": deltavmin,
            "deltavmax": deltavmax,
            "num": n,
            "betastar": beta,
            "gammastar": gamma,
            "taustar": tau,
            "primal": -(n + 1) * (1 + beta + gamma + tau),
            "qfunc": qfunc,
        }

    @staticmethod
    def interval(datagen, wmin, wmax, alpha=0.05, rmin=0, rmax=1, raiseonerr=False):
        from math import inf, isclose, sqrt
        from scipy.stats import f

        assert wmin < 1
        assert wmax > 1
        assert rmin <= rmax

        n, sumw, sumwsq, sumwr, sumwsqr, sumwsqrsq = 0, 0, 0, 0, 0, 0

        for c, w, r in datagen():
            n += c
            sumw += c * w
            sumwsq += c * w ** 2
            sumwr += c * w * r
            sumwsqr += c * w ** 2 * r
            sumwsqrsq += c * w ** 2 * r ** 2
        assert n > 0

        return CrMinusTwo.intervalimpl(
            n,
            sumw,
            sumwsq,
            sumwr,
            sumwsqr,
            sumwsqrsq,
            wmin,
            wmax,
            alpha,
            rmin,
            rmax,
            raiseonerr,
        )

    @staticmethod
    def intervalimpl(
        n,
        sumw,
        sumwsq,
        sumwr,
        sumwsqr,
        sumwsqrsq,
        wmin,
        wmax,
        alpha=0.05,
        rmin=0,
        rmax=1,
        raiseonerr=False,
    ):
        from math import inf, isclose, sqrt
        from scipy.stats import f

        assert wmin < 1
        assert wmax > 1
        assert rmin <= rmax

        uncwfake = wmax if sumw < n else wmin
        if uncwfake == inf:
            uncgstar = 1 + 1 / n
        else:
            unca = (uncwfake + sumw) / (1 + n)
            uncb = (uncwfake ** 2 + sumwsq) / (1 + n)
            uncgstar = (n + 1) * (unca - 1) ** 2 / (uncb - unca * unca)
        Delta = f.isf(q=alpha, dfn=1, dfd=n)
        phi = (-uncgstar - Delta) / (2 * (n + 1))

        bounds = []
        for r, sign in ((rmin, 1), (rmax, -1)):
            candidates = []
            for wfake in (wmin, wmax):
                if wfake == inf:
                    x = sign * (r + (sumwr - sumw * r) / n)
                    y = (r * sumw - sumwr) ** 2 / (n * (1 + n)) - (
                        r ** 2 * sumwsq - 2 * r * sumwsqr + sumwsqrsq
                    ) / (1 + n)
                    z = phi + 1 / (2 * n)
                    if isclose(y * z, 0, abs_tol=1e-9):
                        y = 0

                    if z <= 0 and y * z >= 0:
                        kappa = sqrt(y / (2 * z))
                        if isclose(kappa, 0):
                            candidates.append((sign * r, None))
                        else:
                            gstar = x - sqrt(2 * y * z)
                            gamma = -kappa * (1 + n) / n + sign * (r * sumw - sumwr) / n
                            beta = -sign * r
                            candidates.append(
                                (
                                    gstar,
                                    {
                                        "kappastar": kappa,
                                        "betastar": beta,
                                        "gammastar": gamma,
                                        "wfake": wfake,
                                        # Q_{w,r} &= -\frac{\gamma + \beta w + w r}{(N+1) \kappa} \\
                                        "qfunc": lambda c, w, r, k=kappa, g=gamma, b=beta, s=sign, num=n: -c
                                        * (g + (b + s * r) * w)
                                        / ((num + 1) * k),
                                    },
                                )
                            )
                else:
                    barw = (wfake + sumw) / (1 + n)
                    barwsq = (wfake * wfake + sumwsq) / (1 + n)
                    barwr = sign * (wfake * r + sumwr) / (1 + n)
                    barwsqr = sign * (wfake * wfake * r + sumwsqr) / (1 + n)
                    barwsqrsq = (wfake * wfake * r * r + sumwsqrsq) / (1 + n)

                    if barwsq > barw ** 2:
                        x = barwr + (
                            (1 - barw) * (barwsqr - barw * barwr) / (barwsq - barw ** 2)
                        )
                        y = (barwsqr - barw * barwr) ** 2 / (barwsq - barw ** 2) - (
                            barwsqrsq - barwr ** 2
                        )
                        z = phi + (1 / 2) * (1 - barw) ** 2 / (barwsq - barw ** 2)
                        if isclose(y * z, 0, abs_tol=1e-9):
                            y = 0

                        if z <= 0 and y * z >= 0:
                            kappa = sqrt(y / (2 * z)) if y * z > 0 else 0
                            if isclose(kappa, 0):
                                candidates.append((sign * r, None))
                            else:
                                gstar = x - sqrt(2 * y * z)
                                beta = (
                                    -kappa * (1 - barw) - (barwsqr - barw * barwr)
                                ) / (barwsq - barw * barw)
                                gamma = -kappa - beta * barw - barwr
                                candidates.append(
                                    (
                                        gstar,
                                        {
                                            "kappastar": kappa,
                                            "betastar": beta,
                                            "gammastar": gamma,
                                            "wfake": wfake,
                                            # Q_{w,r} &= -\frac{\gamma + \beta w + w r}{(N+1) \kappa} \\
                                            "qfunc": lambda c, w, r, k=kappa, g=gamma, b=beta, s=sign, num=n: -c
                                            * (g + (b + s * r) * w)
                                            / ((num + 1) * k),
                                        },
                                    )
                                )

            best = min(candidates, key=lambda x: x[0])
            vbound = min(rmax, max(rmin, sign * best[0]))
            bounds.append((vbound, best[1]))

        return (bounds[0][0], bounds[1][0]), (bounds[0][1], bounds[1][1])

    @staticmethod
    def intervaldiff(
        datagen, umin, umax, wmin, wmax, alpha=0.05, rmin=0, rmax=1, raiseonerr=False
    ):
        import numpy as np
        from math import isclose, sqrt
        from scipy.stats import f

        assert umin >= 0
        assert umin < 1
        assert umax > 1
        assert wmin >= 0
        assert wmin < 1
        assert wmax > 1
        assert rmax >= rmin

        _, mle = CrMinusTwo.estimatediff(
            datagen, umin, umax, wmin, wmax, rmin, rmax, raiseonerr=raiseonerr
        )

        Delta = f.isf(q=alpha, dfn=1, dfd=mle["num"] - 1)
        phi = (-Delta - mle["primal"]) / (2 * (mle["num"] + 1))

        n, sumu, sumw, sumuw, sumusq, sumwsq = 0, 0, 0, 0, 0, 0
        sumuMwr, sumuuMwr, sumwuMwr, sumuMwsqrsq = 0, 0, 0, 0
        for c, u, w, r in datagen():
            n += c
            sumu += c * u
            sumw += c * w
            sumuw += c * u * w
            sumusq += c * u ** 2
            sumwsq += c * w ** 2
            sumuMwr += c * (u - w) * r
            sumuuMwr += c * u * (u - w) * r
            sumwuMwr += c * w * (u - w) * r
            sumuMwsqrsq += c * (u - w) ** 2 * r ** 2

        assert n > 0

        bounds = []
        for sign in (1, -1):
            candidates = []
            for ufake, wfake in ((u, w) for u in (umin, umax) for w in (wmin, wmax)):
                rex = rmin if sign * ufake >= sign * wfake else rmax

                baru = (sumu + ufake) / (n + 1)
                barw = (sumw + wfake) / (n + 1)
                barusq = (sumusq + ufake ** 2) / (n + 1)
                barwsq = (sumwsq + wfake ** 2) / (n + 1)
                baruw = (sumuw + ufake * wfake) / (n + 1)
                baruMwr = sign * (sumuMwr + (ufake - wfake) * rex) / (n + 1)
                baruuMwr = sign * (sumuuMwr + ufake * (ufake - wfake) * rex) / (n + 1)
                barwuMwr = sign * (sumwuMwr + wfake * (ufake - wfake) * rex) / (n + 1)
                baruMwsqrsq = (sumuMwsqrsq + (ufake - wfake) ** 2 * rex ** 2) / (n + 1)

                C = np.array(
                    [
                        [1.0, baru, barw],
                        [baru, barusq, baruw],
                        [barw, baruw, barwsq],
                    ],
                    dtype="float64",
                )
                d = np.array([baruMwr, baruuMwr, barwuMwr], dtype="float64")

                a = np.linalg.lstsq(C, np.ones(3), rcond=-1)[0]
                b = np.linalg.lstsq(C, d, rcond=-1)[0]
                x = np.sum(b)
                y = np.dot(d, b) - baruMwsqrsq
                z = phi - 0.5 + 0.5 * np.sum(a)

                if isclose(y * z, 0, abs_tol=1e-9):
                    y = 0

                if z <= 0 and y * z >= 0:
                    gstar = x - sqrt(2 * y * z)
                    kappa = sqrt(y / (2 * z)) if y * z > 0 else 0
                    beta, gamma, tau = -kappa * a - b

                    candidates.append(
                        (
                            gstar,
                            None
                            if isclose(kappa, 0)
                            else {
                                "kappastar": kappa,
                                "betastar": beta,
                                "gammastar": gamma,
                                "taustar": tau,
                                "ufake": ufake,
                                "wfake": wfake,
                                "rfake": rex,
                                "qfunc": lambda c, u, w, r, k=kappa, g=gamma, b=beta, t=tau, s=sign, num=n: -c
                                * (b + g * u + t * w + s * (u - w) * r)
                                / ((num + 1) * k),
                                "mle": mle,
                            },
                        )
                    )

            best = min(candidates, key=lambda x: x[0])
            vbound = min(rmax - rmin, max(rmin - rmax, sign * best[0]))
            bounds.append((vbound, best[1]))

        return (bounds[0][0], bounds[1][0]), (bounds[0][1], bounds[1][1]), candidates
